/* COPYRIGHT NOTICE
 *
 * This program comes from "The X Toolkit Cookbook" by Paul E. Kimball
 * c. 1995 Prentice Hall.
 *
 *  HDR1: Example 8-6 - sortit.c - Alternate input sources
 *  
 *  HDR2: Chapter 8, page 196
 *
 * This software is provided free of charge and may be freely distributed
 * in its original form, so long as this notice appears unchanged in each 
 * copy.  This software is supplied "as-is" and is not covered by warrantee
 * of any kind, express or implied.  
 *
 * This and other examples are available by anonymous ftp from
 * ftp.netcom.com, directory "pub/pk/pkimball/cookbook".
 *
 *  1/15/95	Paul Kimball - pkimball@netcom.com
 */

/*****  Include Required Header Files  *******************************/

#include <sys/wait.h>                /* wait()                       */
#include <X11/Intrinsic.h>           /* The Intrinsics               */
#include <Xm/Xm.h>                   /* Motif resource symbols       */
#include <Xm/Label.h>                /* Motif XmLabel                */
#include <Xm/PushB.h>                /* Motif XmPushButton           */
#include <Xm/Text.h>                 /* Motif XmText                 */
#include <Xm/BulletinB.h>            /* Motif XmBulletinBoard        */

/***** Declare Global Variables  *************************************/

XtAppContext  context;

Widget    toplevel, manager;
Widget    top_button, bottom_button, top_label, text_area;

int       pid;                        /* PID of parent or child       */
char*     readbuffer;                 /* Buffer - data read from child*/
char*     writebuffer;                /* Buffer to write to child     */
int       total_read, total_written;  /* Bytes transfered so far      */
XtInputId read_id, write_id;          /* Input source ID's            */

/***** Forward Declare Callback Routines ******************************/

void ExitCallback();                  /* Exit routine                 */
void ReadPipe();                      /* Reads pipe from child process*/
void WritePipe();                     /* Writes pipe to child process */
void ForkSort();                      /* Creates child process, pipes */
                                      /* and registers input sources  */

/***** Main Program  **************************************************/

main(argc,argv)
int argc;
char **argv;
{

 toplevel = XtAppInitialize(&context, "Sortit",
     NULL, 0, &argc, argv, NULL, NULL, 0);

 manager = XtCreateManagedWidget("manager", 
     xmBulletinBoardWidgetClass, toplevel, NULL, 0);

 top_label = XtCreateManagedWidget("topLabel",
     xmLabelWidgetClass, manager, NULL, 0);

 top_button = XtCreateManagedWidget("topButton",
     xmPushButtonWidgetClass, manager, NULL, 0);

 text_area = XtCreateManagedWidget("textArea",
     xmTextWidgetClass, manager, NULL, 0);

 bottom_button = XtCreateManagedWidget("bottomButton",
     xmPushButtonWidgetClass, manager, NULL, 0);

 XtAddCallback( top_button, XmNactivateCallback,         /* Add      */
     ForkSort, NULL );                                   /* callbacks*/

 XtAddCallback( bottom_button, XmNactivateCallback,
     ExitCallback, NULL );

 XtRealizeWidget(toplevel);             /* Realize all               */
 XtAppMainLoop(context);                /* Start main loop           */
}

/***** Callback Routines *********************************************/

void ForkSort(w, client_data, call_data)       /* Forks child        */
Widget  w;                                     /* Unused             */
XtPointer client_data, call_data;              /* Unused             */
{

  int  status, i;
  int  pipe_to_parent[2], pipe_to_child[2];

  /* Create pipes and try to fork child process */
  pipe(pipe_to_parent);                 /* input pipe from child     */
  pipe(pipe_to_child);                  /* output pipe to child      */

  if (pid = fork(), pid > 0){           /* We're the parent          */
    close(pipe_to_parent[1]);           /* Close child's ends so we  */
    close(pipe_to_child[0]);            /*  don't mess with them     */

    /* Get string to be sorted   */
    XtVaGetValues(text_area, XmNvalue, &writebuffer, NULL);
    total_written = 0;                  /* We've written nothing yet */

    readbuffer = (char*)malloc(0);      /* We've read nothing yet    */
    total_read = 0;

    XtSetSensitive(text_area, False);   /* Don't let user try again  */
    XtSetSensitive(top_button, False);  /*  until we're done         */

    read_id = XtAppAddInput(            /* Register pipe as source   */
      context,                          /* Application context       */
      pipe_to_parent[0],                /* File descriptor - input   */
      XtInputReadMask,                  /* Call when we can read it  */
      ReadPipe,                         /* Routine to call back      */
      NULL);                            /* No client_data required   */

    write_id = XtAppAddInput(           /* Register pipe as source   */
      context,                          /* Application context       */
      pipe_to_child[1],                 /* File descriptor - output  */
      XtInputWriteMask,                 /* Call when we can write it */
      WritePipe,                        /* Routine to call back      */
      NULL);                            /* No client_data required   */

    return;                             /* Back to main loop         */

  }
  else if ( pid == -1 ){                /* We're parent, fork failed */

    XtAppError(context, "Sortit: Sorry - fork failed");

  }
  else if ( pid == 0 ){                 /* We're the child           */

    close(0); dup(pipe_to_child[0]);    /* redirect stdin            */
    close(1); dup(pipe_to_parent[1]);   /* redirect stdout           */

    for(i=2; i<getdtablesize(); i++)    /* Close all other fd's      */
       close(i);          

    execlp("sort", "sort", (char*)0);   /* execute the sort  command */ 
    write(1, "exec failed\0", 12);      /* Oops! - must have failed  */
    _exit(1);                           /* Price of failure = death  */   
  }
}

/*********************************************************************/

void ReadPipe(client_data, source, inputid)
XtPointer client_data;                  /* unused                    */
int       *source;                      /* fd to be read             */
XtInputId *inputid;                     /* Identifies input source   */
{
  union wait status;                    /* status for wait() service */
  int fd = *source;                     /* file to read              */
  char buffer[80];                      /* A buffer to read into     */
  int num_read;                         /* Bytes read this time      */

  num_read = read(fd, buffer, 80);      /* try to read 80 chars      */

  total_read = total_read + num_read; 
  readbuffer = XtRealloc(readbuffer, total_read+1);
  bcopy(buffer, (char*)&readbuffer[total_read-num_read], num_read);

  readbuffer[total_read] = (char)NULL;  /* NULL-terminate the string */

  if(num_read==0) {                     /* End-of-file detected  */

    XtRemoveInput(*inputid);            /* Unregister the source     */
    close(fd);                          /* Now safe to close pipe    */

    pid = wait(&status);                /* Let child die gracefully  */

    /* Display page we just got and prepare for next sort  */
    XtVaSetValues(text_area, XmNvalue, readbuffer, NULL );

    XtFree(readbuffer);                 /* Free input buffer         */
    XtSetSensitive(text_area, True);    /* Set widgets sensitive     */
    XtSetSensitive(top_button, True);   /* so user can try again     */
  }
}

/*********************************************************************/

void WritePipe(client_data, source, inputid)
XtPointer client_data;              /* unused                    */
int       *source;                      /* fd of pipe                */
XtInputId *inputid;                     /* Identifies source         */
{
  int fd = *source;                     /* File to write             */
  int try_write;                        /* Num bytes to try to write */
  int num_written;                      /* Bytes actually written    */
  int num_left;                         /* bytes remaining           */

  num_left = strlen(writebuffer) + 1 - total_written;
  
  if (num_left >= 80) try_write=80;     /* Each time, write 80 chars */
  else try_write = num_left;            /* Last write might be <80   */
     
  /* Write to the pipe         */
  num_written = write(fd,(char*)&writebuffer[total_written], try_write); 
  total_written = total_written + num_written;

  if(total_written == (strlen(writebuffer)+1)) {  /* Finished        */

    write(fd, (char)0x4, 1);                      /* Write EOF       */
  
    XtRemoveInput(*inputid);            /* Remove input source       */
    close(fd);                          /* Safe to close pipe        */
    XtFree(writebuffer);                /* Free buffer               */
  }
}

/*********************************************************************/

void ExitCallback(w, client_data, call_data)     /* Exits            */
Widget  w;                                       /* Unused           */
XtPointer client_data, call_data;                /* Unused           */
{
  exit(1);
}

